/******************************************************************************

	
******************************************************************************/

/************* TWEAKABLES **************/

#define DCC_MAX

string ParamID = "0x000001";
float4 ClearColor : DIFFUSE = {0,0,0,0};
float ClearDepth <string UIWidget = "none";> = 1.0;

float Script : STANDARDSGLOBAL
<
	string UIWidget = "none";
	string ScriptClass = "object";
	string ScriptOrder = "postprocess";
	string ScriptOutput = "color";
	
	// We just call a script in the main technique.
	string Script = "Technique=NvidiaShadowOnly?NvidiaShadowOnly:NvidiaTest:NvidiaTest;";

> = 0.8;

half4 ambient : Ambient
<
    string UIName = "Ambient Color";
> = {0.25f, 0.25f, 0.25f, 1.0f};

half4 surfColor : Diffuse
<
    string UIName = "Diffuse Color";
> = {1.0f, 1.0f, 1.0f, 1.0f};

texture colorTexture : DiffuseMap
<
    string name = "default_color.dds";
	string UIName = "Diffuse Texture";
    string TextureType = "2D";
>;

half4 specularColor : Specular
<
    string UIName = "Specular Color";
> = { 0.2f, 0.2f, 0.2f, 1.0f };

half shininess <
	string UIWidget = "slider";
    float UIMin = 1;
    float UIMax = 128;
    string UIName = "Shininess";
> = 40;

texture normalMap : NormalMap
<
    string name = "default_bump_normal.dds";
	string UIName = "Normal Map";
    string TextureType = "2D";
>;

texture CnormalMap : CNormalMap
<
    string name = "default_bump_normal.dds";
	string UIName = "DXT5 Normal Map";
    string TextureType = "2D";
>;

bool direction
<
    string gui = "slider";
    string UIName = "Flip Normal Map Green Channel";
> = false;

/************** light info **************/

float4 lightPos : POSITION
<
	string UIName = "Light Position";
	string Object = "SpotLight";
	string Space = "World";
	int refID = 0;
> = {100.0f, 100.0f, 100.0f, 0.0f};

float4 light1Color : LIGHTCOLOR
<
	int LightRef = 0;
	string UIName = "Lamp";
	string Object = "SpotLight";
	string UIWidget = "None";
> = { 1.0f, 1.0f, 1.0f, 0.0f };

float SpotLightIntensity <
	string UIName = "Light Intensity";
	string UIWidget = "slider";
	float UIMin = 0.0;
	float UIMax = 2;
	float UIStep = 0.1;
> = 1;

float SpotLightCone <
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 90.5;
    float UIStep = 0.1;
    string UIName = "Cone Angle";
> = 45.0f;

float ShadBias <
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 0.3;
    float UIStep = 0.0001;
    string UIName = "Shadow Bias";
> = 0.01;

half farClip <
	string UIWidget = "slider";
    float UIMin = 0.1;
    float UIMax = 2048;
    string UIName = "Far Clip";
> = 40;

half nearClip <
	string UIWidget = "slider";
    float UIMin = 0.001;
    float UIMax = 2.0;
    string UIName = "Near Clip";
> = 0.001;

/****************************************************/
/********** SAMPLERS ********************************/
/****************************************************/

sampler2D colorTextureSampler = sampler_state
{
	Texture = <colorTexture>;
	MinFilter = Linear;
	MagFilter = Linear;
	MipFilter = Anisotropic;
};

sampler2D normalMapSampler = sampler_state
{
	Texture = <normalMap>;
	MinFilter = Linear;
	MagFilter = Linear;
	MipFilter = Anisotropic;
};

sampler2D CnormalMapSampler = sampler_state
{
	Texture = <CnormalMap>;
	MinFilter = Linear;
	MagFilter = Linear;
	MipFilter = Anisotropic;
};

/***********************************************/
/*** automatically-tracked "tweakables" ********/
/***********************************************/

half4x4 wvp : WorldViewProjection < string UIWidget = "None"; >;
float4x4 world : World < string UIWidget = "None"; >;
half4x4 worldIT : WorldInverseTranspose < string UIWidget = "None"; >;
float4x4 Projection : PROJECTION < string UIWidget = "None"; >;
half4x4 viewInv : ViewInverse < string UIWidget = "None"; >;

float4x4 LampViewXf : View <
    //string UIWidget="None";
    string frustum = "light0";
>;

float4x4 LampProjXf : Projection <
    //string UIWidget="None";
    string frustum = "light0";
>;

///////////////////////////////////////////////////////////
///////////////////////////// Render-to-Texture Data //////
///////////////////////////////////////////////////////////

#define RTT_SIZE 512

texture ShadowMap1 : RENDERCOLORTARGET < 
	float2 Dimensions = { RTT_SIZE, RTT_SIZE };
    int MIPLEVELS = 1;
    string format = "R32F"; //this is a floating point target for the depth map
    //string format = "X8R8G8B8";  // this is regular RGB color
    string UIWidget = "None";
>;

sampler ShadowSamp1 = sampler_state 
{
    texture = <ShadowMap1>;
    AddressU  = CLAMP;
    AddressV = CLAMP;
    MipFilter = NONE;
    MinFilter = POINT;
    MagFilter = POINT;
};

texture ShadDepthTarget : RENDERDEPTHSTENCILTARGET <
	float2 Dimensions = { RTT_SIZE, RTT_SIZE };
    string format = "D24S8";
    string UIWidget = "None";
>;


/****************************************************/
/********** CG SHADER FUNCTIONS *********************/
/****************************************************/

// input from application
struct a2v {
	half4 position		: POSITION;
	half2 texCoord		: TEXCOORD0;
	half3 tangent		: TANGENT;
	half3 binormal		: BINORMAL;
	half3 normal		: NORMAL;
};

// Connector from vertex to pixel shader
struct JustShadowVertexOutput {
    float4 HPosition	: POSITION;
    float4 LP		: TEXCOORD0;	// current position in light-projection space
};

// output to fragment program
struct v2f2 {
        half4 position    	: POSITION;
		half2 texCoord    	: TEXCOORD0;
		half3 eyeVec		: TEXCOORD1;
		half3 lightVec   	: TEXCOORD2;
		half3 worldNormal	: TEXCOORD3;
		half3 worldTangent	: TEXCOORD4;
		half3 worldBinormal	: TEXCOORD5;
		half4 LP			: TEXCOORD6;
};

struct v2fquad {
        half4 position    	: POSITION;
		half2 texCoord    	: TEXCOORD0;
};

//create light transform matrix
float4x4 lightXform(float4 lightPos)
	{
	float3 dirZ = -normalize(lightPos);
	float3 up = float3(0,0,1);
	float3 dirX = normalize(cross(up, dirZ));
	float3 dirY = normalize(cross(dirX, dirZ));
	
	float4x4 lightXfm;
	
	float lightX, lightY, lightZ;
	lightX = -dot(dirX, lightPos);
	lightY = -dot(dirY, lightPos);
	lightZ = -dot(dirZ, lightPos);
	
	float lightW = length(lightPos);
	
	lightXfm[0] = float4(dirX.x, dirY.x, dirZ.x, 0);
	lightXfm[1] = float4(dirX.y, dirY.y, dirZ.y, 0);
	lightXfm[2] = float4(dirX.z, dirY.z, dirZ.z, 0);
	lightXfm[3] = float4(lightX, lightY, lightZ, lightW);
	
	return lightXfm;
	}
	
//create light projection matrix
float4x4 lightProj()
	{
	//these values probably should be exposed to the artist
	float fov = radians(SpotLightCone);
	float near = nearClip;
	float far = farClip;
	
	//compute matrix values
	half halfFov = fov * 0.5;
	half s,c;
	sincos(halfFov,s,c);
	half snear = -s * near;
	half lft = snear / c;
	half rgt = snear / -c;
	half top = rgt;
	half bot = lft;
	
	//build the final matrix
	float4x4 lightProjM = 0;
	lightProjM[0][0] = 2.0 * near / ( rgt - lft );
	lightProjM[1][1] = 2.0 * near / ( top - bot );
	lightProjM[2][2] = far / ( far - near );
	lightProjM[2][3] = 1.0;
	lightProjM[3][2] = near * far / ( near - far );
	lightProjM[2][0] = ( rgt + lft ) / ( rgt - lft );
	lightProjM[2][1] = ( top + bot ) / ( top - bot );
	
	return lightProjM;
	}
	
// blinn lighting with lit function
half4 blinn2(half3 N,
		half3 L,
		half3 V,
		uniform half4 diffuseColor,
		uniform half4 specularColor,
		uniform half shininess
		)
	{
	half3 H = normalize(V+L);
	half4 lighting = lit(dot(L,N), dot(H,N), shininess);
	return diffuseColor*lighting.y + specularColor*lighting.z;
	}



/**************************************/
/***** VERTEX SHADER ******************/
/**************************************/

JustShadowVertexOutput shadVS2(a2v IN)
{
    JustShadowVertexOutput OUT = (JustShadowVertexOutput)0;
    float4 Po = float4(IN.position.xyz,(float)1.0);	// object coordinates
    float4 Pw = mul(Po,world);			// "P" in world coordinates

    //create light transform matrix
	float4x4 lightXfm = lightXform(lightPos);
	
	//create light projection matrix
	float4x4 lightProjM = lightProj();
	
    float4 Pl = mul(Pw,mul(lightXfm,lightProjM));  // "P" in light coords
    OUT.LP = Pl;		// view coords (also lightspace projection coords in this case)
    OUT.HPosition = Pl; // screen clipspace coords
    OUT.LP.z = length(lightPos - Pw);
    return OUT;
}

// from scene camera POV - full featured version
v2f2 mainCamVS(a2v In)
{
    v2f2 Out = (v2f2)0;
    Out.worldNormal = mul(In.normal, worldIT).xyz;
    Out.worldTangent = mul(In.tangent, worldIT).xyz;
    Out.worldBinormal = mul(In.binormal, worldIT).xyz;
    //can use either positive or negative y format normal maps
    //comment out this if statement to save 6 instructions!
  	if (direction == true) Out.worldTangent = -Out.worldTangent;
    half3 worldSpacePos = mul(In.position, world);
    Out.lightVec = (lightPos - worldSpacePos);		// world coords
    Out.texCoord.xy = In.texCoord;						// pass-thru
    Out.eyeVec = viewInv[3].xyz - worldSpacePos;
    Out.position = mul(In.position, wvp);
    
    //create light transform matrix
	float4x4 lightXfm = lightXform(lightPos);
	
	//create light projection matrix
	float4x4 lightProjM = lightProj();
	
    float4 Pl = mul(worldSpacePos,mul(lightXfm,lightProjM));  // "P" in light coords
    Out.LP = Pl;							// ...for pixel-shader shadow calcs
	Out.lightVec = (lightPos - worldSpacePos);		// world coords
	
    //object position in object space
    //float4 objectSpaceObjectPosition = float4(0,0,0,1);
    //float4 worldSpaceObjectPosition = mul(objectSpaceObjectPosition, world);
    //float4 lightToObjectVector = (lightPos - worldSpaceObjectPosition);
    
    Out.LP.w = length(Out.lightVec);
	
    return Out;
}

// from scene camera POV - stripped down version
v2f2 mainCamVS2(a2v In)
{
    v2f2 Out = (v2f2)0;
    half3 worldSpacePos = mul(In.position, world);
    Out.position = mul(In.position, wvp);
    
    //create light transform matrix
	float4x4 lightXfm = lightXform(lightPos);
	
	//create light projection matrix
	float4x4 lightProjM = lightProj();
	
    float4 Pl = mul(worldSpacePos,mul(lightXfm,lightProjM));  // "P" in light coords
    Out.LP = Pl;							// ...for pixel-shader shadow calcs
    Out.lightVec = (lightPos - worldSpacePos);		// world coords
    
    //object position in object space
    //float4 objectSpaceObjectPosition = float4(0,0,0,1);
    //float4 worldSpaceObjectPosition = mul(objectSpaceObjectPosition, world);
    //float4 lightToObjectVector = (lightPos - worldSpaceObjectPosition);
    
    Out.LP.w = length(Out.lightVec);
    return Out;
}

//for writing a quad to the screen with just the buffer texture applied
v2fquad VS_Quad(float3 Position : POSITION, float3 TexCoord : TEXCOORD0)
{
    v2fquad OUT = (v2fquad)0;
    OUT.position = float4(Position, 1);
    OUT.texCoord = float4(TexCoord, 1); 
    return OUT;
}

/**************************************/
/***** FRAGMENT PROGRAM ***************/
/**************************************/

// puts a floating point value into a 4 channel texture
float4 pack(float depth) 
{
	float v = depth / farClip;
	v *= 16777216.0f;
    float rf = v / 65536.0f;
	float r = rf - frac( rf );
	v -= r * 65536.0f;
	float gf = v / 256.0f;
	float g = gf - frac( gf );
	v -= g * 256.0f;
	float b = v;
	r /= 256.0f;
	g /= 256.0f;
	b /= 256.0f;

	return float4(r,g,b,1.0);
}

//converts a 4 channel texture into one floating point value
float unpack(float4 f)
{
	float v = ( ( f.r * 16777216.0f ) + ( f.g * 65536.0f ) + ( f.b * 256.0f ) );
	v /= 16777216.0f;
	v *= farClip;
	return v;
}


//
// Utility function for pixel shaders to use this shadow map
//
float shadow_calc(float4 LP, uniform sampler ShadowMapSampler, uniform float ShadowBiasing) 
{
    float2 nuv = float2(0.5,-0.5)*LP.xy/LP.w + float2(0.5,0.5);
	float4 temp = tex2D(ShadowMapSampler,nuv);
    float shadMapDepth = unpack(temp);
    
    float depth = LP.w - ShadowBiasing;
    float shad = 1-(shadMapDepth<depth);
	return shad;
}

float4 shadPS(JustShadowVertexOutput IN) : COLOR
{
	return pack(IN.LP.z); 
}

float4 useShadowPS(v2f2 In, uniform float CosSpotAng) : COLOR
{
    float3 Nn = normalize(In.worldNormal);
    float3 Vn = normalize(In.eyeVec);
    float3 Ln = normalize(In.lightVec);
    float3 Hn = normalize(Vn + Ln);
    float hdn = dot(Hn,Nn);
    float ldn = dot(Ln,Nn);
    float4 litVec = lit(ldn,hdn,shininess);
    ldn = litVec.y * SpotLightIntensity;
    float dl = -normalize(In.LP.xyz).z;
    dl = max((float)0,((dl-CosSpotAng)/(((float)1.0)-CosSpotAng)));
    float3 ambiContrib = surfColor * ambient;
    float3 diffContrib = surfColor*(ldn * light1Color);
    float3 specContrib = ((ldn * litVec.z * specularColor) * light1Color);
    float3 result = diffContrib + specContrib;
    float shadowed = shadow_calc(In.LP,ShadowSamp1,ShadBias);
    return float4((dl*shadowed*result)+ambiContrib,1);
}

//stripped down version
float4 useShadowPS2(v2f2 In, uniform float CosSpotAng) : COLOR
{
    float shadowed = shadow_calc(In.LP,ShadowSamp1,ShadBias);
    return float4(shadowed,shadowed,shadowed,1);
}

//for writing a quad to the screen with just the buffer texture applied
float4 PS_TestPass(v2fquad IN) : COLOR
{   
	return tex2D(ShadowSamp1, float2(IN.texCoord.x, IN.texCoord.y));
}

/****************************************************/
/********** TECHNIQUES ******************************/
/****************************************************/

technique NvidiaShadowOnly <
	string Script = "Pass=MakeShadow;"
					"Pass=UseShadow;";
> {
	pass MakeShadow <
		string Script = "RenderColorTarget0=ShadowMap1;"
						"RenderDepthStencilTarget=ShadDepthTarget;"
						"RenderPort=light0;"
						"ClearSetColor=ShadowClearColor;"
						"ClearSetDepth=ClearDepth;"
						"Clear=Color;"
						"Clear=Depth;"
						"Draw=geometry;";
	> {
		VertexShader = compile vs_1_1 shadVS2();
		ZEnable = true;
		ZWriteEnable = true;
		ZFunc = LessEqual;
		CullMode = none;
		PixelShader = compile ps_2_0 shadPS();
	}
	pass UseShadow <
		string Script = "RenderColorTarget0=;"
						"RenderDepthStencilTarget=;"
						"RenderPort=;"
						//"ClearSetColor=ClearColor;"
						//"ClearSetDepth=ClearDepth;"
						//"Clear=Color;"
						//"Clear=Depth;"
						"Draw=geometry;";
	> {
		VertexShader = compile vs_1_1 mainCamVS2();
		ZEnable = true;
		ZWriteEnable = true;
		ZFunc = LessEqual;
		CullMode = None;
		PixelShader = compile ps_2_0 useShadowPS2(cos(radians(SpotLightCone)));
	}
}

technique NvidiaTest <
	string Script = "Pass=MakeShadow;"
					"Pass=Screen;";
> {
	pass MakeShadow <
		string Script = "RenderColorTarget0=ShadowMap1;"
						"RenderDepthStencilTarget=ShadDepthTarget;"
						"RenderPort=light0;"
						"ClearSetColor=ShadowClearColor;"
						"ClearSetDepth=ClearDepth;"
						"Clear=Color;"
						"Clear=Depth;"
						"Draw=geometry;";
	> {
		VertexShader = compile vs_1_1 shadVS2();
		ZEnable = true;
		ZWriteEnable = true;
		ZFunc = LessEqual;
		CullMode = None;
		PixelShader = compile ps_2_0 shadPS();
	}
	pass Screen
    <
       	string Script= "RenderColorTarget0=;"
						"Draw=Buffer;";        	
	>
    {
		cullmode = none;
		ZEnable = true;
		ZWriteEnable = true;
		AlphaBlendEnable = false;
		VertexShader = compile vs_1_1 VS_Quad();
		PixelShader = compile ps_1_1 PS_TestPass();	
    }
}

technique Nvidia <
	string Script = "Pass=MakeShadow;"
					"Pass=UseShadow;";
> {
	pass MakeShadow <
		string Script = "RenderColorTarget0=ShadowMap1;"
						"RenderDepthStencilTarget=ShadDepthTarget;"
						"RenderPort=light0;"
						"ClearSetColor=ShadowClearColor;"
						"ClearSetDepth=ClearDepth;"
						"Clear=Color;"
						"Clear=Depth;"
						"Draw=geometry;";
	> {
		VertexShader = compile vs_1_1 shadVS2();
		ZEnable = true;
		ZWriteEnable = true;
		ZFunc = LessEqual;
		CullMode = None;
		PixelShader = compile ps_2_0 shadPS();
	}
	pass UseShadow <
		string Script = "RenderColorTarget0=;"
						"RenderDepthStencilTarget=;"
						"RenderPort=;"
						//"ClearSetColor=ClearColor;"
						//"ClearSetDepth=ClearDepth;"
						//"Clear=Color;"
						//"Clear=Depth;"
						"Draw=geometry;";
	> {
		VertexShader = compile vs_1_1 mainCamVS();
		ZEnable = true;
		ZWriteEnable = true;
		ZFunc = LessEqual;
		CullMode = None;
		PixelShader = compile ps_2_0 useShadowPS(cos(radians(SpotLightCone)));
	}
}

